import EventEmitter from 'tiny-emitter';
import extend from 'extend';
import throttle from 'lodash.throttle';
import $ from './el';
import scrollHandler from './scroll-handler';
import Pagination from './pagination';
import Debugger from './debugger';
import nextHandler from './next-handler';
import {getRootRect, getBoundingClientRect, getScroll} from './dimensions';
import Spinner from './spinner';

const defaults = {
  item: '',
  pagination: {},
  next: '',
  responseType: 'document',
  bind: true,
  scrollContainer: window,
  spinner: false,
};

let loggers = {
  hit: function() {
    console.log('Hit scroll threshold');
  },
  binded: function() {
    console.log('Binded event handlers');
  },
  unbinded: function() {
    console.log('Unbinded event handlers');
  },
  // scrolled: function() {
  //   console.log('Scrolled');
  // },
  next: function(event) {
    console.log(`Next page triggered [pageIndex=${event.pageIndex}]`);
  },
  load: function(url, xhr) {
    console.log(`Start loading ${url}`);
  },
  loaded: function(url, xhr) {
    console.log(`Finished loading`);
  },
  append: function() {
    console.log('Start appending new items');
  },
  appended: function() {
    console.log('Finished appending new items');
  },
  noneLeft: function() {
    console.log('No more pages left to load');
  },
};

let scrollListener;
let resizeListener;

export default class IAS {
  constructor(container, options) {
    this.eventEmitter = new EventEmitter();
    this.options = extend({}, defaults, options);

    Object.keys(loggers).forEach((key) => {
      this.on(key, loggers[key]);
    });

    this.container = $(container).get();
    this.scrollContainer = $(this.options.scrollContainer).get();

    if (!this.scrollContainer) {
      throw new Error(`Scroll container ${scrollContainer} not found`);
    }

    this.pagination = new Pagination(this, this.options.pagination);
    this.spinner = new Spinner(this, this.options.spinner);

    this.nextHandler = nextHandler;
    if (typeof this.options.next === 'function') {
      this.nextHandler = this.options.next;
    }

    this.pageIndex = 0;

    this.on('hit', this.next); // @todo find better name?

    if (this.options.bind) {
      // @todo on document.ready?
      this.bind();
    }
  }

  bind() {
    scrollListener = throttle(scrollHandler, 200).bind(this);
    resizeListener = throttle(this.measure, 200).bind(this);

    this.scrollContainer.addEventListener('scroll', scrollListener);
    this.scrollContainer.addEventListener('resize', resizeListener);

    this.eventEmitter.emit('binded');

    this.measure();
  }

  unbind() {
    this.scrollContainer.removeEventListener('scroll', scrollListener);
    this.scrollContainer.removeEventListener('resize', resizeListener);

    this.eventEmitter.emit('unbinded');
  }

  next() {
    this.pause();

    let event = {
      pageIndex: this.pageIndex,
    };

    this.eventEmitter.emit('next', event);

    Promise.resolve(this.nextHandler(event.pageIndex))
        .then((result) => {
          if (!result) {
            this.eventEmitter.emit('noneLeft');

            return;
          }

          this.pageIndex++;
          this.resume();
        })
    ;
  }

  pause() {
    this.paused = true;
  }

  resume() {
    this.paused = false;

    this.measure();
  }

  load(url) {
    let ias = this;

    return new Promise((resolve, reject) => {
      let xhr = new XMLHttpRequest();

      xhr.onreadystatechange = function() {
        if (xhr.readyState === XMLHttpRequest.DONE) {
          if (xhr.status === 200) {
            let items = [];

            if (ias.options.responseType === 'document') {
               items = $(ias.options.item, xhr.response).all();
            }

            // @todo define event variable and pass that around

            ias.eventEmitter.emit('loaded', items, url, xhr);

            resolve({items: items, url: url, xhr: xhr});
          } else {
            console.error('Request failed');

            reject(xhr);
          }
        }
      };

      // FIXME: make no-caching configurable
      // @see https://developer.mozilla.org/nl/docs/Web/API/XMLHttpRequest/Using_XMLHttpRequest#Bypassing_the_cache
      url = url + ((/\?/).test(url) ? "&" : "?") + (new Date()).getTime();

      xhr.open('GET', url, true);
      xhr.responseType = ias.options.responseType;
      xhr.setRequestHeader('X-Requested-With', 'XMLHttpRequest');

      ias.eventEmitter.emit('load', url, xhr);

      xhr.send();
    });
  }

  /**
   * @param {array<Element>} items
   * @param {Element|null} parent
   */
  append(items, parent) {
    let ias = this;

    // @todo move fragment creation into executor?
    let insert = document.createDocumentFragment();

    items.forEach((item) => {
      insert.appendChild(item);
    });

    let executor = (resolve) => {
      window.requestAnimationFrame(() => {
        let last = ias.lastElement();
        parent = parent || ias.container;
        let sibling = last ? last.nextSibling : null;

        parent.insertBefore(insert, sibling);

        resolve({items: items, parent: parent});

        ias.eventEmitter.emit('appended', items, parent);
      });
    };

    let event = {
      items: items,
      parent: parent,
      executor: executor,
    };

    ias.eventEmitter.emit('append', event);

    return new Promise(event.executor);
  }

  lastElement() {
    return $(this.options.item, this.container).last();
  }

  measure() {
    let scroll = getScroll(this.scrollContainer);
    let rootRect = getRootRect(this.scrollContainer);
    let boundingRect = getBoundingClientRect(this.lastElement());

    let scrollYBottom = scroll.y + rootRect.height;
    let bottom = scroll.y + boundingRect.bottom - rootRect.top;

    let info = {
      scroll: scroll,
      rootRect: rootRect,
      boundingRect: boundingRect,

      scrollYBottom: scrollYBottom,
      bottom: bottom,
      distanceToFold: bottom - scrollYBottom,
    };

    if (!this.paused && info.distanceToFold <= 0) {
      this.eventEmitter.emit('hit', info); // @todo review values
    }
  }

  on(event, callback) {
    this.eventEmitter.on(event, callback.bind(this));
  }

  off(event, callback) {
    this.eventEmitter.off(event, callback.bind(this));
  }

  once(event, callback) {
    this.eventEmitter.once(event, callback.bind(this));
  }

  debug() {
    this.debugger = (new Debugger(this)).bind();
  }
}
